/***************************************************************************
 *   Copyright (C) 2009 by Stanson <stanson@pangolin.ru>                   *
 *   GPL v3 code, read COPYING file                                        *
 *                                                                         *
 * Ethernet 10BASE-T protocol                                              *
 *                                                                         *
 * Bit time 100 ns                                                         *
 *                                                                         *
 * Logical 0 - Hi to Low                                                   *
 * Logical 1 - Low to Hi                                                   *
 *                                                                         *
 * Ethernet packet structure                                               *
 *                                                                         *
 * [     Preamble    ]  7 bytes 0x55                                       *
 * [       SFD       ]  1 byte 0xD5 Start of Frame Delimiter               *
 * [ Destination MAC ]  6 bytes of destination MAC                         *
 * [    Source MAC   ]  6 bytes of source MAC                              *
 * [  Type / Length  ]  2 bytes of type or length                          *
 * [     Payload     ]  46 or more bytes of data                           *
 * [       FCS       ]  4 bytes of CRC32, 31 bit sent first                *
 *                                                                         *
 * After packet goes TD_IDL pulse - Hi for 300 ns                          *
 *                                                                         *
 * EUSART in syncronous master mode output data to DT/CK pins              *
 * Baudrate is set to 10MHz                                                *
 * using XOR gate on this pins we can get required Manchester encoding     *
 *                                                                         *
 * To keep link alive issue Normal Link Pulse (Hi for 100ns) every 8-24 ms *
 *                                                                         *
 **************************************************************************/

#include <pic18fregs.h>
#include "main.h"
#include "ethernet_tx.h"

#define SET_TMR1_100nS(x) { TMR1H = HIGH(-(x)); TMR1L = LOW(-(x)); }

#pragma udata bank0 etx_frame
ethernet_frame	etx_frame;
byte		etx_cnt;
byte		etx_nlp;

const byte 	etx_header[] =
{
    0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0xD5,	/* preamble */
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05,			/* destination MAC */
    0x00, 0xDE, 0xAD, 0xBE, 0xEF, 0x00,			/* source MAC */
    0x08, 0x00,						/* EtherType (IPv4) */

    /* IPv4 header */
    0x45,						/* Version, Length */
    0x00,						/* Type of Service */
    0x00, ETX_PAYLOAD_SIZE,				/* IP data length */
    0x00, 0x00,						/* Packet counter */
    0x00,						/* IP flags, fragment offset */
    0x00,						/* IP offset */
    0x40,						/* Time To Live */
    0x11,						/* IP protocol (UDP) */
    0x00, 0x00,						/* IP header checksum */
    192, 168, 255, 200,					/* Source IP */
    192, 168, 255, 1,					/* Destination IP */

    /* UDP header */
    0x04, 0x00,						/* Source UDP port */
    0x04, 0x00,						/* Destination UDP port */
    0x00, ETX_UDP_DATAGRAM_SIZE,			/* UDP datagram size */
    0x00, 0x00						/* UDP pseudoheader checksum */
};

void etx_crc32( void )
{
    unsigned long       crc;
    byte                i, j;

    crc = ~0UL; /* First 32 bits of frame is complemented */

    for (i = sizeof(etx_frame.pre); i < sizeof(etx_frame) - sizeof(etx_frame.crc); i++)
    {
	crc ^= etx_frame.raw[i];
	for (j = 0; j < 8; j++)
	{
	    crc >>= 1;
	    if(STATUSbits.C) crc ^= 0xEDB88320;
	}
    }

    etx_frame.crc = ~crc; /* result is complemented */
}

void etx_ip_checksum( void )
{
    unsigned long sum;
    word w;
    byte i;

    etx_frame.payload[10] = 0;
    etx_frame.payload[11] = 0;

    for( sum = 0, i = 0; i < ETX_IP_HEADER_SIZE; i+=2 )
    {
	w = etx_frame.payload[ i ];
	w <<= 8;
	w |= etx_frame.payload[ i + 1 ];
	sum += w;
    }

    while( w = sum >> 16 )
    {
	sum = ( sum & 0xFFFF ) + w;
    }

    sum = ~sum;

    etx_frame.payload[10] = ( sum >> 8 ) & 0xFF;
    etx_frame.payload[11] = sum & 0xFF;
}

void etx_init( void )
{
    byte i;

    /* EUSART init */
    BAUDCTL = 0x00;	/* 10MHz baudrate */
    SPBRGH  = 0x00;
    SPBRG   = 0x00;

    TXSTA   = 0xB0;	/* master, 8-bit, TX enabled, synchronous */
    RCSTA   = 0x00;

    for( i = 0; i < sizeof(etx_frame.payload); i++ )
    {
	etx_frame.payload[i] = 0;
    }

    /* fill header */
    for( i = 0; i < sizeof(etx_header); i++ )
    {
	etx_frame.raw[i] = etx_header[i];
    }

    /* timer2 for NLP */
    T1CON = 0x81; /* Fosc/4 clock 100ns, 1:1 prescaler, enabled */
    SET_TMR1_100nS( 10000 - 22 ); /* 1000 uS period */

    IPR1bits.TMR1IP = 0;
    PIR1bits.TMR1IF = 0;
    PIE1bits.TMR1IE = 1;

    ETX_E = 0;

    TRISB &= ETX_TRISB;

    etx_nlp = ETX_NLP_PERIOD;
}

void etx_send_data( void )
{
    etx_ip_checksum();
    etx_crc32();

    INTCONbits.GIE = 0;

    /* prepare RB1 and RB4 for Hi signal after turning off EUSART */
    ETX_CLK  = 0;
    ETX_DATA = 1;

    /* send frame */
    etx_cnt = sizeof(etx_frame) - 1;

    __asm		LFSR	0, _etx_frame		__endasm;
    /* turn on EUSART */
    __asm		BSF	_RCSTA, 7 	__endasm;
    /* and feed EUSART with bytes of frame */
    __asm 		MOVF	_POSTINC0, W	__endasm;
    __asm 		MOVWF	_TXREG		__endasm;
    __asm		BSF	_PORTB, 0	__endasm; /* 1 cycle */

    __asm _etx_ram:	NOP			__endasm; /* 1 cycle   [ b0 ] <---------+      */
    __asm		NOP			__endasm; /* 1 cycle   [ b1 ]           |      */
    __asm		NOP			__endasm; /* 1 cycle   [ b2 ]           |      */
    __asm 		MOVF	_POSTINC0, W	__endasm; /* 1 cycle   [ b3 ]           |      */
    __asm 		MOVWF	_TXREG		__endasm; /* 1 cycle   [ b4 ]           |      */
    __asm		DECF	_etx_cnt, F, B	__endasm; /* 1 cycle   [ b5 ]           |      */
    __asm		BNZ	_etx_ram	__endasm; /* 1/2 cycle [ b6 ] <if !0> [ b7 ] */
    __asm		NOP			__endasm; /* 1 cycle   [ b7 ] */
    /* wait while last byte come out and turn off EUSART */
    __asm		NOP			__endasm; /* 1 cycle   [ b0 ] */
    __asm		NOP			__endasm; /* 1 cycle   [ b1 ] */
    __asm		NOP			__endasm; /* 1 cycle   [ b2 ] */
    __asm		NOP			__endasm; /* 1 cycle   [ b3 ] */
    __asm		NOP			__endasm; /* 1 cycle   [ b4 ] */
    __asm		NOP			__endasm; /* 1 cycle   [ b5 ] */
    __asm		NOP			__endasm; /* 1 cycle   [ b6 ] */
    __asm		BCF	_RCSTA, 7 	__endasm; /* 1 cycle   [ b7 ] */
    /* generate TP_IDL (300ns of Hi) */
    __asm		NOP			__endasm; /* 1 cycle */
    __asm		NOP			__endasm; /* 1 cycle */
    __asm		BCF	_PORTB, 0	__endasm; /* 1 cycle */

    PIR1bits.TMR1IF = 0;

    etx_nlp = ETX_NLP_PERIOD;

    INTCONbits.GIE = 1;

}

void etx_process( void )
{
    /* send 100ns Normal Link Pulse every 16 ms */
    SET_TMR1_100nS( 10000 - 22 ); /* 1000 uS period */

    PIR1bits.TMR1IF = 0;

    if( --etx_nlp ) return;

    ETX_CLK  = 0;
    ETX_DATA = 1;
    __asm		BSF	_PORTB, 0		__endasm; /* 1 cycle */
    __asm		BCF	_PORTB, 0		__endasm; /* 1 cycle */

    etx_nlp = ETX_NLP_PERIOD;

}
